import sys
from difflib import context_diff
from filecmp import dircmp
from pathlib import Path

import pytest
from typer.testing import CliRunner

from orchestrator.cli.database import app as db_app
from orchestrator.cli.generate import app as generate_app


def absolute_path(path: str) -> str:
    file = Path(__file__).resolve().parent / "data" / path
    return str(file)


def create_main():
    with open("main.py", "w") as fp:
        fp.write(
            "from orchestrator import OrchestratorCore\n"
            "from orchestrator.cli.main import app as core_cli\n"
            "from orchestrator.settings import AppSettings\n"
            "\n"
            "app = OrchestratorCore(base_settings=AppSettings())\n"
            'if __name__ == "__main__":\n'
            "    core_cli()\n"
        )


@pytest.fixture(scope="module")
def monkey_module():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="module")
def actual_folder(tmp_path_factory, monkey_module) -> Path:
    tmp_path = tmp_path_factory.mktemp("generate")
    monkey_module.chdir(tmp_path)
    sys.path.append(str(tmp_path))
    create_main()
    runner = CliRunner()
    runner.invoke(db_app, ["init"])
    for config_file in (absolute_path("product_config2.yaml"), absolute_path("product_config1.yaml")):
        for cmd in ("product-blocks", "product", "workflows", "unit-tests"):
            runner.invoke(generate_app, [cmd, "--config-file", config_file, "--no-dryrun", "--force"])
        runner.invoke(generate_app, ["migration", "--config-file", config_file])
    return tmp_path


@pytest.fixture(scope="module")
def expected_folder() -> Path:
    return Path(__file__).resolve().parent / "data" / "generate"


def test_missing_or_extra_files(expected_folder, actual_folder):
    def assert_equal_dirs(dcmp):
        left_only = ", ".join(dcmp.left_only)
        right_only = ", ".join(dcmp.right_only)
        relative_folder = Path(dcmp.right).relative_to(actual_folder)

        assert left_only == "", f"missing file(s) {left_only} from directory {relative_folder}"
        assert right_only == "", f"extra file(s) {right_only} in directory {relative_folder}"

        for sub_dcmp in dcmp.subdirs.values():
            assert_equal_dirs(sub_dcmp)

    assert_equal_dirs(dircmp(expected_folder, actual_folder, ignore=["versions", "__pycache__"]))  # skip migrations


def test_differences_in_generated_code(expected_folder, actual_folder):
    for expected in expected_folder.rglob("[!0-9]*.py"):  # skip migration versions for now (contain date and uniq id)
        relative = expected.relative_to(expected_folder)
        actual = actual_folder / relative
        diff = context_diff(
            open(expected).readlines(), open(actual).readlines(), fromfile="expected", tofile="actual", n=1, lineterm=""
        )

        assert list(diff) == [], f'generated {relative} differs (see "pytest -vv")'



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements websocket endpoints."""

from typing import Annotated

import structlog
from fastapi import Header, status
from fastapi.routing import APIRouter
from fastapi.websockets import WebSocket

from orchestrator.settings import app_settings
from orchestrator.websocket import WS_CHANNELS, websocket_manager

logger = structlog.get_logger(__name__)

router = APIRouter()

if app_settings.ENABLE_WEBSOCKETS:

    @router.websocket("/events")
    async def websocket_events(
        websocket: WebSocket, sec_websocket_protocol: Annotated[str | None, Header()] = None
    ) -> None:
        """Emits events for the frontend.

        Events are of the form {"name": <name>, "value": <value>}.
        To authenticate, provide a JWT token in the "Sec-Websocket-Protocol" header.
        """
        error = await websocket_manager.authorize(websocket, sec_websocket_protocol or "")

        await websocket.accept()
        if error:
            await websocket_manager.disconnect(websocket, reason=error, code=status.WS_1008_POLICY_VIOLATION)
            return

        await websocket_manager.connect(websocket, WS_CHANNELS.EVENTS)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Generator
from pathlib import Path

import structlog
from more_itertools import first

from orchestrator.cli.generator.generator.enums import to_dict
from orchestrator.cli.generator.generator.settings import product_generator_settings as settings
from orchestrator.utils.helpers import camel_to_snake, snake_to_camel

logger = structlog.getLogger(__name__)


def get_workflow(config: dict, workflow_name: str) -> dict:
    workflows = (workflow for workflow in config.get("workflows", []) if workflow["name"] == workflow_name)
    return first(workflows, {})


def get_variable(config: dict) -> str:
    return config.get("variable", camel_to_snake(config["name"]))


def get_product_block_variable(product_block: dict) -> str:
    return get_variable(product_block)


def get_product_file_name(config: dict) -> str:
    return get_variable(config)


def get_product_block_file_name(product_block: dict) -> str:
    return get_product_block_variable(product_block)


def root_product_block(config: dict) -> dict:
    product_blocks = config.get("product_blocks", [])
    # TODO: multiple product_blocks will need more logic, ok for now
    return product_blocks[0]


def insert_into_imports(content: list[str], new_import: str) -> list[str]:
    # Note: we may consider using a real Python parser here someday, but for now this is ok and formatting
    # gets done by isort and black.
    def produce() -> Generator:
        not_inserted_yet = True
        for line in content:
            if line.startswith("from ") and not_inserted_yet:
                yield new_import
                not_inserted_yet = False
            yield line

    return list(produce())


def path_to_module(path: Path) -> str:
    return str(path).replace("/", ".")


def get_workflows_folder() -> Path:
    return settings.FOLDER_PREFIX / settings.WORKFLOWS_PATH


def get_product_blocks_folder() -> Path:
    return settings.FOLDER_PREFIX / settings.PRODUCT_BLOCKS_PATH


def get_product_types_folder() -> Path:
    return settings.FOLDER_PREFIX / settings.PRODUCT_TYPES_PATH


def get_workflows_module() -> str:
    return path_to_module(get_workflows_folder())


def get_product_blocks_module() -> str:
    return path_to_module(get_product_blocks_folder())


def get_product_types_module() -> str:
    return path_to_module(get_product_types_folder())


def get_product_path(config: dict) -> Path:
    file_name = get_product_file_name(config)
    return get_product_types_folder() / Path(file_name).with_suffix(".py")


def get_product_import(product: dict, lifecycle: str = "") -> str:
    return f'from {get_product_types_module()}.{product["variable"]} import {product["type"]}{lifecycle}\n'


def create_dunder_init_files(path: Path) -> None:
    folder = Path("")
    for part in path.parts:
        if (folder := folder / part).is_dir():
            if not (dunder_init_file := folder / Path("__init__.py")).exists():
                logger.info("creating missing dunder init", path=str(dunder_init_file))
                open(dunder_init_file, "x").close()


def is_constrained_int(field: dict) -> bool:
    return "min_value" in field or "max_value" in field


def get_constrained_ints(fields: list[dict]) -> list[dict]:
    return [field for field in fields if is_constrained_int(field)]


def merge_fields(fields: list[dict], int_enums: list[dict], str_enums: list[dict]) -> list[dict]:
    return list((to_dict(fields) | to_dict(int_enums) | to_dict(str_enums)).values())


def is_name_spaced_field_type(field: dict) -> bool:
    return "." in field["type"]


def name_space_get_type(name_spaced_type: str) -> str:
    return name_spaced_type.split(".")[-1]


def process_fields(fields: list[dict]) -> list[dict]:
    def to_type(field: dict) -> dict:
        if is_constrained_int(field):
            return field | {"type": snake_to_camel(field["name"])}

        if is_name_spaced_field_type(field):
            return field | {"type": name_space_get_type(field["type"])}

        return field

    return [to_type(field) for field in fields]


def get_all_fields(product_block: dict) -> list[dict]:
    return process_fields(product_block["fields"])


def get_input_fields(product_block: dict) -> list[dict]:
    def supported_input_type(field: dict) -> bool:
        return field["type"] in ("int", "str", "bool", "enum")

    return process_fields([field for field in product_block["fields"] if supported_input_type(field)])


def get_name_spaced_types_to_import(fields: list) -> list[tuple]:
    # NOTE: we could make this smarter by grouping imports from the namespace, but isort will handle this for us
    def name_space_split(field: dict) -> tuple[str, str]:
        *namespace, type = field["type"].split(".")
        return ".".join(namespace), type

    return [name_space_split(field) for field in fields if is_name_spaced_field_type(field)]
